@@ -1,6 +1,7 @@
 #!/usr/bin/env python
-import sys, os, datetime, time, shutil, tempfile, subprocess, string
+import sys, os, datetime, time, shutil, tempfile, subprocess
 from optparse import OptionParser
+from ._backup import *
 
 info = "'rethinkdb restore' loads data into a RethinkDB cluster from an archive"
 usage = "rethinkdb restore FILE [-c HOST:PORT] [-a AUTH_KEY] [--clients NUM] [--force] [-i (DB | DB.TABLE)]..."
@@ -18,6 +19,8 @@ def print_restore_help():
     print "                                   be specified multiple times)"
     print "  --clients NUM_CLIENTS            the number of client connections to use (defaults"
     print "                                   to 8)"
+    print "  --hard-durability                use hard durability writes (slower, but less memory"
+    print "                                   consumption on the server)"
     print "  --force                          import data even if a table already exists"
     print ""
     print "EXAMPLES:"
@@ -43,6 +46,7 @@ def parse_options():
     parser.add_option("-a", "--auth", dest="auth_key", metavar="KEY", default="", type="string")
     parser.add_option("-i", "--import", dest="tables", metavar="DB | DB.TABLE", default=[], action="append", type="string")
     parser.add_option("--clients", dest="clients", metavar="NUM_CLIENTS", default=8, type="int")
+    parser.add_option("--hard-durability", dest="hard", action="store_true", default=False)
     parser.add_option("--force", dest="force", action="store_true", default=False)
     parser.add_option("--debug", dest="debug", default=False, action="store_true")
     parser.add_option("-h", "--help", dest="help", default=False, action="store_true")
@@ -61,12 +65,7 @@ def parse_options():
     res = { }
 
     # Verify valid host:port --connect option
-    host_port = options.host.split(":")
-    if len(host_port) == 1:
-        host_port = (host_port[0], "28015") # If just a host, use the default port
-    if len(host_port) != 2:
-        raise RuntimeError("Error: Invalid 'host:port' format: %s" % options.host)
-    (res["host"], res["port"]) = host_port
+    (res["host"], res["port"]) = parse_connect_option(options.host)
 
     # Verify valid input file
     res["in_file"] = os.path.abspath(args[0])
@@ -75,21 +74,11 @@ def parse_options():
         raise RuntimeError("Error: Archive file does not exist: %s" % res["in_file"])
 
     # Verify valid --import options
-    res["dbs"] = []
-    res["tables"] = []
-    for item in options.tables:
-        if not all(c in string.ascii_letters + string.digits + "._" for c in item):
-            raise RuntimeError("Error: Invalid 'db' or 'db.table' name: %s" % item)
-        db_table = item.split(".")
-        if len(db_table) == 1:
-            res["dbs"].append(db_table[0])
-        elif len(db_table) == 2:
-            res["tables"].append(tuple(db_table))
-        else:
-            raise RuntimeError("Error: Invalid 'db' or 'db.table' format: %s" % item)
+    res["tables"] = parse_db_table_options(options.tables)
 
     res["auth_key"] = options.auth_key
     res["clients"] = options.clients
+    res["hard"] = options.hard
     res["force"] = options.force
     res["debug"] = options.debug
     return res
@@ -106,10 +95,13 @@ def do_unzip(temp_dir, options):
 
     # Only untar the selected tables
     # Apparently tar has problems if the same file is specified by two filters, so remove dupes
-    for db in set(options["dbs"]):
-        tar_args.append(os.path.join("*", db))
+    dbs_to_export = set()
     for db, table in set(options["tables"]):
-        if db not in options["dbs"]:
+        if table is None:
+            dbs_to_export.add(db)
+            tar_args.append(os.path.join("*", db))
+    for db, table in set(options["tables"]):
+        if table is not None and db not in dbs_to_export:
             tar_args.append(os.path.join("*", db, table + ".*"))
 
     res = subprocess.call(tar_args)
@@ -127,16 +119,20 @@ def do_import(temp_dir, options):
     import_args.extend(["--auth", options["auth_key"]])
     import_args.extend(["--clients", str(options["clients"])])
 
-    for db in options["dbs"]:
-        import_args.extend(["--import", db])
     for db, table in options["tables"]:
-        import_args.extend(["--import", "%s.%s" % (db, table)])
+        if table is None:
+            import_args.extend(["--import", db])
+        else:
+            import_args.extend(["--import", "%s.%s" % (db, table)])
 
+    if options["hard"]:
+        import_args.append("--hard-durability")
     if options["force"]:
         import_args.append("--force")
     if options["debug"]:
         export_args.extend(["--debug"])
 
+    print "importing with args: %s" % str(import_args)
     res = subprocess.call(import_args)
     if res != 0:
         raise RuntimeError("Error: rethinkdb-import failed")
