@@ -183,16 +183,16 @@ class RqlQuery(object):
         return Mod(other, self)
 
     def __and__(self, other):
-        return All(self, other)
+        return All(self, other, infix=True)
 
     def __rand__(self, other):
-        return All(other, self)
+        return All(other, self, infix=True)
 
     def __or__(self, other):
-        return Any(self, other)
+        return Any(self, other, infix=True)
 
     def __ror__(self, other):
-        return Any(other, self)
+        return Any(other, self, infix=True)
 
     # Non-operator versions of the above
 
@@ -229,6 +229,12 @@ class RqlQuery(object):
     def mod(self, other):
         return Mod(self, other)
 
+    def and_(*args):
+        return All(*args)
+
+    def or_(*args):
+        return Any(*args)
+
     def not_(self):
         return Not(self)
 
@@ -273,11 +279,14 @@ class RqlQuery(object):
     def coerce_to(self, other_type):
         return CoerceTo(self, other_type)
 
+    def ungroup(self):
+        return Ungroup(self)
+
     def type_of(self):
         return TypeOf(self)
 
-    def merge(self, other):
-        return Merge(self, other)
+    def merge(self, *others):
+        return Merge(self, *map(func_wrap, others))
 
     def append(self, val):
         return Append(self, val)
@@ -305,13 +314,22 @@ class RqlQuery(object):
     def __getitem__(self, index):
         if isinstance(index, slice):
             if index.stop:
-                return Slice(self, index.start or 0, index.stop)
+                return Slice(self, index.start or 0, index.stop, bracket_operator=True)
             else:
-                return Slice(self, index.start or 0, -1, right_bound='closed')
+                return Slice(self, index.start or 0, -1, right_bound='closed', bracket_operator=True)
         elif isinstance(index, int):
-            return Nth(self, index)
+            return Nth(self, index, bracket_operator=True)
         elif isinstance(index, types.StringTypes):
-            return GetField(self, index)
+            return GetField(self, index, bracket_operator=True)
+        elif isinstance(index, RqlQuery):
+            raise RqlDriverError(
+                "Bracket operator called with a ReQL expression parameter.\n"+
+                "Dynamic typing is not supported in this syntax,\n"+
+                "use `.nth`, `.slice`, or `.get_field` instead.")
+        else:
+            raise RqlDriverError(
+                "bracket operator called with an unsupported parameter type: %s.%s" %
+                (index.__class__.__module__, index.__class__.__name__))
 
     def __iter__(self):
         raise RqlDriverError(
@@ -328,6 +346,15 @@ class RqlQuery(object):
     def match(self, pattern):
         return Match(self, pattern)
 
+    def split(self, *args):
+        return Split(self, *args)
+
+    def upcase(self):
+        return Upcase(self)
+
+    def downcase(self):
+        return Downcase(self)
+
     def is_empty(self):
         return IsEmpty(self)
 
@@ -343,8 +370,20 @@ class RqlQuery(object):
     def limit(self, index):
         return Limit(self, index)
 
-    def reduce(self, func, base=()):
-        return Reduce(self, func_wrap(func), base=base)
+    def reduce(self, func):
+        return Reduce(self, func_wrap(func))
+
+    def sum(self, *args):
+        return Sum(self, *[func_wrap(arg) for arg in args])
+
+    def avg(self, *args):
+        return Avg(self, *[func_wrap(arg) for arg in args])
+
+    def min(self, *args):
+        return Min(self, *[func_wrap(arg) for arg in args])
+
+    def max(self, *args):
+        return Max(self, *[func_wrap(arg) for arg in args])
 
     def map(self, func):
         return Map(self, func_wrap(func))
@@ -368,7 +407,7 @@ class RqlQuery(object):
     # NB: Can't overload __len__ because Python doesn't
     #     allow us to return a non-integer
     def count(self, filter=()):
-        if filter == ():
+        if filter is ():
             return Count(self)
         else:
             return Count(self, func_wrap(filter))
@@ -383,18 +422,13 @@ class RqlQuery(object):
         return OuterJoin(self, other, predicate)
 
     def eq_join(self, left_attr, other, index=()):
-        return EqJoin(self, left_attr, other, index=index)
+        return EqJoin(self, func_wrap(left_attr), other, index=index)
 
     def zip(self):
         return Zip(self)
 
-    def grouped_map_reduce(self, grouping, mapping, data_collector, base=()):
-        return GroupedMapReduce(self, func_wrap(grouping), func_wrap(mapping),
-            func_wrap(data_collector), base=base)
-
-    def group_by(self, arg1, arg2, *rest):
-        args = [arg1, arg2] + list(rest)
-        return GroupBy(self, list(args[:-1]), args[-1])
+    def group(self, *args, **kwargs):
+        return Group(self, *[func_wrap(arg) for arg in args], **kwargs)
 
     def for_each(self, mapping):
         return ForEach(self, func_wrap(mapping))
@@ -470,12 +504,44 @@ class RqlQuery(object):
 def needs_wrap(arg):
     return isinstance(arg, Datum) or isinstance(arg, MakeArray) or isinstance(arg, MakeObj)
 
+class RqlBoolOperQuery(RqlQuery):
+    def __init__(self, *args, **optargs):
+        if 'infix' in optargs:
+            self.infix = optargs['infix']
+            del optargs['infix']
+        else:
+            self.infix = False
+
+        RqlQuery.__init__(self, *args, **optargs)
+
+    def compose(self, args, optargs):
+        t_args = [T('r.expr(', args[i], ')') if needs_wrap(self.args[i]) else args[i] for i in xrange(len(args))]
+
+        if self.infix:
+            return T('(', T(*t_args, intsp=[' ', self.st_infix, ' ']), ')')
+        else:
+            return T('r.', self.st, '(', T(*t_args, intsp=', '), ')')
+
 class RqlBiOperQuery(RqlQuery):
     def compose(self, args, optargs):
-        if needs_wrap(self.args[0]) and needs_wrap(self.args[1]):
-            args[0] = T('r.expr(', args[0], ')')
+        t_args = [T('r.expr(', args[i], ')') if needs_wrap(self.args[i]) else args[i] for i in xrange(len(args))]
+        return T('(', T(*t_args, intsp=[' ', self.st, ' ']), ')')
 
-        return T('(', args[0], ' ', self.st, ' ', args[1], ')')
+class RqlBiCompareOperQuery(RqlBiOperQuery):
+    def __init__(self, *args, **optargs):
+        RqlBiOperQuery.__init__(self, *args, **optargs)
+
+        for arg in args:
+            try:
+                if arg.infix:
+                    err = "Calling '%s' on result of infix bitwise operator:\n" + \
+                          "%s.\n" + \
+                          "This is almost always a precedence error.\n" + \
+                          "Note that `a < b | b < c` <==> `a < (b | b) < c`.\n" + \
+                          "If you really want this behavior, use `.or_` or `.and_` instead."
+                    raise RqlDriverError(err % (self.st, QueryPrinter(self).print_query()))
+            except AttributeError:
+                pass # No infix attribute, so not possible to be an infix bool operator
 
 class RqlTopLevelQuery(RqlQuery):
     def compose(self, args, optargs):
@@ -493,6 +559,24 @@ class RqlMethodQuery(RqlQuery):
 
         return T(args[0], '.', self.st, '(', restargs, ')')
 
+class RqlBracketQuery(RqlMethodQuery):
+    def __init__(self, *args, **optargs):
+        if 'bracket_operator' in optargs:
+            self.bracket_operator = optargs['bracket_operator']
+            del optargs['bracket_operator']
+        else:
+            self.bracket_operator = False
+
+        RqlMethodQuery.__init__(self, *args, **optargs)
+
+    def compose(self, args, optargs):
+        if self.bracket_operator:
+            if needs_wrap(self.args[0]):
+                args[0] = T('r.expr(', args[0], ')')
+            return T(args[0], '[', T(*args[1:], intsp=[',']), ']')
+        else:
+            return RqlMethodQuery.compose(self, args, optargs)
+
 class RqlTzinfo(datetime.tzinfo):
 
     def __init__(self, offsetstr):
@@ -525,6 +609,24 @@ def reql_type_time_to_datetime(obj):
     else:
         return datetime.datetime.utcfromtimestamp(obj['epoch_time'])
 
+# Python only allows immutable built-in types to be hashed, such as for keys in a dict
+# This means we can't use lists or dicts as keys in grouped data objects, so we convert
+# them to tuples and frozensets, respectively.
+# This may make it a little harder for users to work with converted grouped data, unless
+# they do a simple iteration over the result
+def recursively_make_hashable(obj):
+    if isinstance(obj, list):
+        return tuple([recursively_make_hashable(i) for i in obj])
+    elif isinstance(obj, dict):
+        return frozenset([(k, recursively_make_hashable(v)) for (k,v) in obj.items()])
+    return obj
+
+def reql_type_grouped_data_to_object(obj):
+    if not 'data' in obj:
+        raise RqlDriverError('pseudo-type GROUPED_DATA object %s does not have the expected field "data".' % py_json.dumps(obj))
+
+    return dict([(recursively_make_hashable(k),v) for (k,v) in obj['data']])
+
 # This class handles the conversion of RQL terminal types in both directions
 # Going to the server though it does not support R_ARRAY or R_OBJECT as those
 # are alternately handled by the MakeArray and MakeObject nodes. Why do this?
@@ -560,51 +662,58 @@ class Datum(RqlQuery):
         return repr(self.data)
 
     @staticmethod
-    def _convert_pseudotype(obj, time_format):
+    def _convert_pseudotype(obj, format_opts):
         reql_type = obj.get('$reql_type$')
         if reql_type is not None:
             if reql_type == 'TIME':
-                if time_format == 'native':
+                time_format = format_opts.get('time_format')
+                if time_format is None or time_format == 'native':
                     # Convert to native python datetime object
                     return reql_type_time_to_datetime(obj)
                 elif time_format != 'raw':
                     raise RqlDriverError("Unknown time_format run option \"%s\"." % time_format)
+            elif reql_type == 'GROUPED_DATA':
+                group_format = format_opts.get('group_format')
+                if group_format is None or group_format == 'native':
+                    return reql_type_grouped_data_to_object(obj)
+                elif group_format != 'raw':
+                    raise RqlDriverError("Unknown group_format run option \"%s\"." % group_format)
             else:
                 raise RqlDriverError("Unknown pseudo-type %s" % reql_type)
         # If there was no pseudotype, or the time format is raw, return the original object
         return obj
 
     @staticmethod
-    def _recursively_convert_pseudotypes(obj, time_format):
+    def _recursively_convert_pseudotypes(obj, format_opts):
         if isinstance(obj, dict):
             for (key, value) in obj.iteritems():
-                obj[key] = Datum._recursively_convert_pseudotypes(value, time_format)
-            obj = Datum._convert_pseudotype(obj, time_format)
+                obj[key] = Datum._recursively_convert_pseudotypes(value, format_opts)
+            obj = Datum._convert_pseudotype(obj, format_opts)
         elif isinstance(obj, list):
             for i in xrange(len(obj)):
-                obj[i] = Datum._recursively_convert_pseudotypes(obj[i], time_format)
+                obj[i] = Datum._recursively_convert_pseudotypes(obj[i], format_opts)
         return obj
 
     @staticmethod
-    def deconstruct(datum, time_format='native'):
+    def deconstruct(datum, format_opts={}):
         d_type = datum.type
         if d_type == p.Datum.R_JSON:
             obj = py_json.loads(datum.r_str)
-            return Datum._recursively_convert_pseudotypes(obj, time_format)
+            return Datum._recursively_convert_pseudotypes(obj, format_opts)
         elif d_type == p.Datum.R_OBJECT:
             obj = { }
             for pair in datum.r_object:
-                obj[pair.key] = Datum.deconstruct(pair.val, time_format)
+                obj[pair.key] = Datum.deconstruct(pair.val, format_opts)
 
             # Thanks to "pseudo-types" we can't yet be quite sure if this object is meant to
             # be an object or something else. We need a second layer of type switching, this
             # time on an obfuscated field "$reql_type$" rather than the datum type field we
             # already switched on.
-            Datum._convert_pseudotype(obj, time_format)
+            Datum._convert_pseudotype(obj, format_opts)
             return obj
         elif d_type == p.Datum.R_ARRAY:
             array = datum.r_array
-            return [Datum.deconstruct(e, time_format) for e in array]
+            return [Datum.deconstruct(e, format_opts) for e in array]
         elif d_type == p.Datum.R_STR:
             return datum.r_str
         elif d_type == p.Datum.R_NUM:
@@ -675,27 +784,27 @@ class ImplicitVar(RqlQuery):
     def compose(self, args, optargs):
         return 'r.row'
 
-class Eq(RqlBiOperQuery):
+class Eq(RqlBiCompareOperQuery):
     tt = p.Term.EQ
     st = "=="
 
-class Ne(RqlBiOperQuery):
+class Ne(RqlBiCompareOperQuery):
     tt = p.Term.NE
     st = "!="
 
-class Lt(RqlBiOperQuery):
+class Lt(RqlBiCompareOperQuery):
     tt = p.Term.LT
     st = "<"
 
-class Le(RqlBiOperQuery):
+class Le(RqlBiCompareOperQuery):
     tt = p.Term.LE
     st = "<="
 
-class Gt(RqlBiOperQuery):
+class Gt(RqlBiCompareOperQuery):
     tt = p.Term.GT
     st = ">"
 
-class Ge(RqlBiOperQuery):
+class Ge(RqlBiCompareOperQuery):
     tt = p.Term.GE
     st = ">="
 
@@ -755,11 +864,18 @@ class SetDifference(RqlMethodQuery):
     tt = p.Term.SET_DIFFERENCE
     st = "set_difference"
 
-class Slice(RqlQuery):
+class Slice(RqlBracketQuery):
     tt = p.Term.SLICE
+    st = 'slice'
 
+    # Slice has a special bracket syntax, implemented here
     def compose(self, args, optargs):
-        return T(args[0], '[', args[1], ':', args[2], ']')
+        if self.bracket_operator:
+            if needs_wrap(self.args[0]):
+                args[0] = T('r.expr(', args[0], ')')
+            return T(args[0], '[', args[1], ':', args[2], ']')
+        else:
+            return RqlBracketQuery.compose(self, args, optargs)
 
 class Skip(RqlMethodQuery):
     tt = p.Term.SKIP
@@ -769,11 +885,9 @@ class Limit(RqlMethodQuery):
     tt = p.Term.LIMIT
     st = 'limit'
 
-class GetField(RqlQuery):
+class GetField(RqlBracketQuery):
     tt = p.Term.GET_FIELD
-
-    def compose(self, args, optargs):
-        return T(args[0], '[', args[1], ']')
+    st = 'get_field'
 
 class Contains(RqlMethodQuery):
     tt = p.Term.CONTAINS
@@ -791,6 +905,10 @@ class Keys(RqlMethodQuery):
     tt = p.Term.KEYS
     st = 'keys'
 
+class Object(RqlMethodQuery):
+    tt = p.Term.OBJECT
+    st = 'object'
+
 class Pluck(RqlMethodQuery):
     tt = p.Term.PLUCK
     st = 'pluck'
@@ -814,8 +932,8 @@ class DB(RqlTopLevelQuery):
     def table_list(self):
         return TableList(self)
 
-    def table_create(self, table_name, primary_key=(), datacenter=(), cache_size=(), durability=()):
-        return TableCreate(self, table_name, primary_key=primary_key, datacenter=datacenter, cache_size=cache_size, durability=durability)
+    def table_create(self, table_name, primary_key=(), datacenter=(), durability=()):
+        return TableCreate(self, table_name, primary_key=primary_key, datacenter=datacenter, durability=durability)
 
     def table_drop(self, table_name):
         return TableDrop(self, table_name)
@@ -887,6 +1005,22 @@ class Reduce(RqlMethodQuery):
     tt = p.Term.REDUCE
     st = 'reduce'
 
+class Sum(RqlMethodQuery):
+    tt = p.Term.SUM
+    st = 'sum'
+
+class Avg(RqlMethodQuery):
+    tt = p.Term.AVG
+    st = 'avg'
+
+class Min(RqlMethodQuery):
+    tt = p.Term.MIN
+    st = 'min'
+
+class Max(RqlMethodQuery):
+    tt = p.Term.MAX
+    st = 'max'
+
 class Map(RqlMethodQuery):
     tt = p.Term.MAP
     st = 'map'
@@ -915,16 +1049,26 @@ class Union(RqlMethodQuery):
     tt = p.Term.UNION
     st = 'union'
 
-class Nth(RqlQuery):
+class Nth(RqlBracketQuery):
     tt = p.Term.NTH
-
-    def compose(self, args, optargs):
-        return T(args[0], '[', args[1], ']')
+    st = 'nth'
 
 class Match(RqlMethodQuery):
     tt = p.Term.MATCH
     st = 'match'
 
+class Split(RqlMethodQuery):
+    tt = p.Term.SPLIT
+    st = 'split'
+
+class Upcase(RqlMethodQuery):
+    tt = p.Term.UPCASE
+    st = 'upcase'
+
+class Downcase(RqlMethodQuery):
+    tt = p.Term.DOWNCASE
+    st = 'downcase'
+
 class IndexesOf(RqlMethodQuery):
     tt = p.Term.INDEXES_OF
     st = 'indexes_of'
@@ -933,13 +1077,9 @@ class IsEmpty(RqlMethodQuery):
     tt = p.Term.IS_EMPTY
     st = 'is_empty'
 
-class GroupedMapReduce(RqlMethodQuery):
-    tt = p.Term.GROUPED_MAP_REDUCE
-    st = 'grouped_map_reduce'
-
-class GroupBy(RqlMethodQuery):
-    tt = p.Term.GROUPBY
-    st = 'group_by'
+class Group(RqlMethodQuery):
+    tt = p.Term.GROUP
+    st = 'group'
 
 class InnerJoin(RqlMethodQuery):
     tt = p.Term.INNER_JOIN
@@ -961,6 +1101,10 @@ class CoerceTo(RqlMethodQuery):
     tt = p.Term.COERCE_TO
     st = 'coerce_to'
 
+class Ungroup(RqlMethodQuery):
+    tt = p.Term.UNGROUP
+    st = 'ungroup'
+
 class TypeOf(RqlMethodQuery):
     tt = p.Term.TYPEOF
     st = 'type_of'
@@ -1045,13 +1189,15 @@ class Branch(RqlTopLevelQuery):
     tt = p.Term.BRANCH
     st = "branch"
 
-class Any(RqlBiOperQuery):
+class Any(RqlBoolOperQuery):
     tt = p.Term.ANY
-    st = "|"
+    st = "or_"
+    st_infix = "|"
 
-class All(RqlBiOperQuery):
+class All(RqlBoolOperQuery):
     tt = p.Term.ALL
-    st = "&"
+    st = "and_"
+    st_infix = "&"
 
 class ForEach(RqlMethodQuery):
     tt = p.Term.FOREACH
